/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.reports.networkDesign;

import com.net2plan.interfaces.networkDesign.IReport;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.internal.ForwardingRuleComparator;
import com.net2plan.libraries.GraphTheoryMetrics;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.utils.*;
import com.net2plan.utils.Constants.RoutingType;
import java.awt.geom.Point2D;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import org.codehaus.stax2.XMLOutputFactory2;
import org.codehaus.stax2.XMLStreamWriter2;

/**
 * This report shows basic information about the network design.
 * 
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class Report_networkDesign implements IReport
{
	@Override
	public String executeReport(NetPlan netPlan, Map<String, String> reportParameters, Map<String, String> net2planParameters)
	{
		try (ByteArrayOutputStream os = new ByteArrayOutputStream())
		{
			XMLOutputFactory2 output = (XMLOutputFactory2) XMLOutputFactory.newFactory();
			XMLStreamWriter2 writer = (XMLStreamWriter2) output.createXMLStreamWriter(os);
			
			writer.writeStartDocument("UTF-8", "1.0");
			writer.writeStartElement("network");
			
			int N = netPlan.getNumberOfNodes();
			Set<Long> nodeIds = netPlan.getNodeIds();
			
			Set<Long> srgIds = netPlan.getSRGIds();

			Set<Long> layerIds = netPlan.getLayerIds();
			for(long layerId : layerIds)
			{
				writer.writeStartElement("layer");
				writer.writeAttribute("id", Long.toString(layerId));
				writer.writeAttribute("name", netPlan.getLayerName(layerId));
				
				int E = netPlan.getNumberOfLinks(layerId);
				int D = netPlan.getNumberOfDemands(layerId);

				double PRECISION_FACTOR = Double.parseDouble(net2planParameters.get("precisionFactor"));

				double averageNodeDegree = N == 0 ? 0 : (double) E / (double) N;
				Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap();
				double averageLinkLength = 0;
				double maxCapacity = 0;
				double minCapacity = 0;
				double averageCapacity = 0;
				double totalCapacity = 0;
				int networkDiameter_hops = 0;
				double networkDiameter_km = 0;
				double networkDiameter_ms = 0;
				double averageShortestPath_hops = 0;
				double averageShortestPath_km = 0;
				double averageShortestPath_ms = 0;
				double capacityModule = 0;

				boolean isConnected = false;
				boolean isBidirectional = false;
				boolean isWeightedBidirectional = false;
				boolean isSimple = false;

				double Y_e = netPlan.getLinkTotalCarriedTraffic(layerId);

				if (E > 0)
				{
					Map<Long, Double> d_e = netPlan.getLinkLengthInKmMap(layerId);
					Map<Long, Double> u_e = netPlan.getLinkCapacityMap(layerId);
					Map<Long, Double> t_e_prop = netPlan.getLinkPropagationDelayInSecondsMap(layerId);

					isConnected = GraphUtils.isConnected(nodeIds, linkMap);
					isBidirectional = GraphUtils.isBidirectional(linkMap);
					isWeightedBidirectional = GraphUtils.isWeightedBidirectional(linkMap, u_e);
					isSimple = GraphUtils.isSimple(linkMap);

					averageLinkLength = DoubleUtils.sum(d_e);
					double[] maxMinCapacity = DoubleUtils.maxMinValues(u_e);
					maxCapacity = maxMinCapacity[0];
					minCapacity = maxMinCapacity[1];
					averageCapacity = DoubleUtils.average(u_e);

					GraphTheoryMetrics metrics = new GraphTheoryMetrics(nodeIds, linkMap, null);
					networkDiameter_hops = (int) metrics.getDiameter();
					averageShortestPath_hops = metrics.getAverageShortestPathDistance();

					metrics.configureLinkCostMap(d_e);
					networkDiameter_km = metrics.getDiameter();
					averageShortestPath_km = metrics.getAverageShortestPathDistance();

					metrics.configureLinkCostMap(t_e_prop);
					networkDiameter_ms = 1000 * metrics.getDiameter();
					averageShortestPath_ms = 1000 * metrics.getAverageShortestPathDistance();

					capacityModule = DoubleUtils.gcd(u_e.values());

					if (capacityModule < PRECISION_FACTOR) capacityModule = 0;
				}
				
				RoutingType routingType_thisLayer = netPlan.getRoutingType(layerId);
				writer.writeAttribute("routingType", routingType_thisLayer == RoutingType.HOP_BY_HOP_ROUTING ? "hopByHopRouting" : "sourceRouting");
				
				/* Put topology metrics */
				writer.writeAttribute("topologyMetrics_N", String.format("%d", N));
				writer.writeAttribute("topologyMetrics_E", String.format("%d", E));
				writer.writeAttribute("topologyMetrics_averageNodeDegree", String.format("%.3f", averageNodeDegree));
				writer.writeAttribute("topologyMetrics_averageLinkLength", String.format("%.3f", averageLinkLength));
				writer.writeAttribute("topologyMetrics_networkDiameter_hops", String.format("%d", networkDiameter_hops));
				writer.writeAttribute("topologyMetrics_networkDiameter_km", String.format("%.3f", networkDiameter_km));
				writer.writeAttribute("topologyMetrics_networkDiameter_ms", String.format("%.3f", networkDiameter_ms));
				writer.writeAttribute("topologyMetrics_averageShortestPath_hops", String.format("%.3f", averageShortestPath_hops));
				writer.writeAttribute("topologyMetrics_averageShortestPath_km", String.format("%.3f", averageShortestPath_km));
				writer.writeAttribute("topologyMetrics_averageShortestPath_ms", String.format("%.3f", averageShortestPath_ms));
				writer.writeAttribute("topologyMetrics_isConnected", String.format("%s", isConnected ? "Yes" : "No"));
				writer.writeAttribute("topologyMetrics_isBidirectional", String.format("%s", isBidirectional ? "Yes" : "No"));
				writer.writeAttribute("topologyMetrics_isWeightedBidirectional", String.format("%s", isWeightedBidirectional ? "Yes" : "No"));
				writer.writeAttribute("topologyMetrics_isSimple", String.format("%s", isSimple ? "Yes" : "No"));

				/* Put capacity metrics */
				writer.writeAttribute("capacityMetrics_maxCapacity", String.format("%.3f", maxCapacity));
				writer.writeAttribute("capacityMetrics_minCapacity", String.format("%.3f", minCapacity));
				writer.writeAttribute("capacityMetrics_averageCapacity", String.format("%.3f", averageCapacity));
				writer.writeAttribute("capacityMetrics_totalCapacity", String.format("%.3f", totalCapacity));
				writer.writeAttribute("capacityMetrics_capacityModule", String.format("%.3f", capacityModule));

				double H_d = netPlan.getDemandTotalOfferedTraffic();
				double R_d = netPlan.getDemandTotalCarriedTraffic();
				double B_d = netPlan.getDemandTotalBlockedTraffic();
				writer.writeAttribute("trafficMetrics_totalOfferedTraffic", String.format("%.3f", H_d));
				writer.writeAttribute("trafficMetrics_totalCarriedTraffic", String.format("%.3f", R_d));
				writer.writeAttribute("trafficMetrics_averageOfferedTraffic", String.format("%.3f", D == 0 ? 0 : H_d / D));
				writer.writeAttribute("trafficMetrics_averageCarriedTraffic", String.format("%.3f", D == 0 ? 0 : R_d / D));
				writer.writeAttribute("trafficMetrics_blockedTraffic", String.format("%.3f", H_d == 0 ? 0 : Math.max(100 * B_d / H_d, 0)));
				writer.writeAttribute("trafficMetrics_numHops", String.format("%.3f", R_d == 0 ? 0 : Y_e / R_d));
				
				/* Per-node information */
				for(long nodeId : nodeIds)
				{
					Point2D nodeXYPosition = netPlan.getNodeXYPosition(nodeId);
					Set<Long> srgIds_thisNode = netPlan.getNodeSRGs(nodeId);
					double ingressTraffic = netPlan.getNodeIngressTraffic(layerId, nodeId);
					double egressTraffic = netPlan.getNodeEgressTraffic(layerId, nodeId);
					double traversingTraffic = netPlan.getNodeTraversingTraffic(layerId, nodeId);
					Map<String, String> nodeAttributes = netPlan.getNodeAttributeMap(nodeId);

					writer.writeStartElement("node");
					writer.writeAttribute("id", Long.toString(nodeId));
					writer.writeAttribute("name", Long.toString(nodeId));
					writer.writeAttribute("isDown", netPlan.isNodeDown(nodeId) ? "x" : "");
					writer.writeAttribute("xCoord", String.format("%.3f", nodeXYPosition.getX()));
					writer.writeAttribute("yCoord", String.format("%.3f", nodeXYPosition.getY()));
					writer.writeAttribute("ingressTraffic", String.format("%.3f", ingressTraffic));
					writer.writeAttribute("egressTraffic", String.format("%.3f", egressTraffic));
					writer.writeAttribute("traversingTraffic", String.format("%.3f", traversingTraffic));
					writer.writeAttribute("totalTraffic", String.format("%.3f", ingressTraffic + egressTraffic + traversingTraffic));
					writer.writeAttribute("srgIds", srgIds_thisNode.isEmpty() ? "none" : CollectionUtils.join(srgIds_thisNode, ", "));
					writer.writeAttribute("attributes", nodeAttributes.isEmpty() ? "none" : StringUtils.mapToString(nodeAttributes));
					writer.writeEndElement();
				}
				
				/* Per-link information */
				Set<Long> linkIds = netPlan.getLinkIds(layerId);
				double max_rho_e = netPlan.getLinkMaximumUtilization(layerId);
				for(long linkId : linkIds)
				{
					Pair<Long, Long> nodePair_thisLink = netPlan.getLinkOriginDestinationNodePair(layerId, linkId);
					long originNodeId = nodePair_thisLink.getFirst();
					long destinationNodeId = nodePair_thisLink.getSecond();
					String originNodeName = netPlan.getNodeName(originNodeId);
					String destinationNodeName = netPlan.getNodeName(destinationNodeId);
					
					Set<Long> srgIds_thisLink = netPlan.getLinkSRGs(layerId, linkId);
					Pair<Long, Long> coupledDemand = netPlan.isLinkCoupled(layerId, linkId) ? netPlan.getLinkCoupledLowerLayerDemand(layerId, linkId) : null;

					double rho_e = netPlan.getLinkUtilization(layerId, linkId);
					writer.writeStartElement("link");
					writer.writeAttribute("id", Long.toString(linkId));
					writer.writeAttribute("originNodeId", originNodeId + (originNodeName.isEmpty() ? "" : " (" + originNodeName + ")"));
					writer.writeAttribute("destinationNodeId", destinationNodeId + (destinationNodeName.isEmpty() ? "" : " (" + destinationNodeName + ")"));
					writer.writeAttribute("isDown", netPlan.isLinkDown(layerId, linkId) ? "x" : "");
					writer.writeAttribute("coupledTo", coupledDemand == null ? "" : String.format("demand %d (layer %d)", coupledDemand.getSecond(), coupledDemand.getFirst()));
					writer.writeAttribute("capacity", String.format("%.3f", netPlan.getLinkCapacity(layerId, linkId)));
					writer.writeAttribute("carriedTraffic", String.format("%.3f", netPlan.getLinkCarriedTraffic(layerId, linkId)));
					writer.writeAttribute("reservedCapacity", String.format("%.3f", netPlan.getLinkCapacityReservedForProtection(layerId, linkId)));
					writer.writeAttribute("utilization", String.format("%.3f", netPlan.getLinkUtilization(layerId, linkId)));
					writer.writeAttribute("utilizationWithoutProtection", String.format("%.3f", netPlan.getLinkUtilizationWithoutConsiderReservedCapacityForProtection(layerId, linkId)));
					writer.writeAttribute("isBottleneck", DoubleUtils.isEqualWithinRelativeTolerance(max_rho_e, rho_e, PRECISION_FACTOR) == true ? "x" : "");
					writer.writeAttribute("lengthInKm", String.format("%.3f", netPlan.getLinkLengthInKm(layerId, linkId)));
					writer.writeAttribute("propagationSpeedInKmPerSecond", String.format("%.3f", netPlan.getLinkPropagationSpeedInKmPerSecond(layerId, linkId)));
					writer.writeAttribute("propagationDelayInMiliseconds", String.format("%.3f", 1000 * netPlan.getLinkPropagationDelayInSeconds(layerId, linkId)));

					if (routingType_thisLayer == RoutingType.SOURCE_ROUTING)
					{
						Set<Long> traversingRoutes = netPlan.getLinkAssociatedRoutes(layerId, linkId);
						Set<Long> traversingSegments = netPlan.getLinkAssociatedProtectionSegments(layerId, linkId);
						int numRoutes = traversingRoutes.size();
						int numSegments = traversingSegments.size();
						String routesString = numRoutes + (numRoutes > 0 ? " (" + CollectionUtils.join(traversingRoutes, ", ") + ")" : "");
						String segmentsString = numSegments + (numSegments > 0 ? " (" + CollectionUtils.join(traversingSegments, ", ") + ")" : "");
						writer.writeAttribute("routeIds", routesString);
						writer.writeAttribute("segmentIds", segmentsString);
					}
					else
					{
						Set<Pair<Long, Long>> forwardingRules = netPlan.getLinkForwardingRules(layerId, linkId);
						int numForwardingRules = forwardingRules.size();

						StringBuilder forwardingRulesString = new StringBuilder(Integer.toString(numForwardingRules));
						if (numForwardingRules > 0)
						{
							forwardingRulesString.append(" (");
							Iterator<Pair<Long, Long>> it = forwardingRules.iterator();
							while(it.hasNext())
							{
								Pair<Long, Long> demandLinkPair = it.next();
								forwardingRulesString.append("d").append(demandLinkPair.getFirst()).append(" : ").append(String.format("%.3f", netPlan.getForwardingRuleSplittingRatio(layerId, demandLinkPair)));
								if (it.hasNext()) forwardingRulesString.append(", ");
							}

							forwardingRulesString.append(")");
						}

						writer.writeAttribute("forwardingRules", forwardingRulesString.toString());
					}

					writer.writeAttribute("srgIds", srgIds_thisLink.isEmpty() ? "none" : CollectionUtils.join(srgIds_thisLink, ", "));

					Map<String, String> linkAttributes = netPlan.getLinkAttributeMap(layerId, linkId);
					writer.writeAttribute("attributes", linkAttributes.isEmpty() ? "none" : StringUtils.mapToString(linkAttributes));
					writer.writeEndElement();
				}					
				
				/* Per-demand information */
				Set<Long> demandIds = netPlan.getDemandIds(layerId);
				for(long demandId : demandIds)
				{
					Pair<Long, Long> coupledLink = netPlan.isDemandCoupled(layerId, demandId) ? netPlan.getDemandCoupledUpperLayerLink(layerId, demandId) : null;
					double h_d = netPlan.getDemandOfferedTraffic(layerId, demandId);
					double lostTraffic_d = netPlan.getDemandBlockedTraffic(layerId, demandId);
					Set<Long> routeIds_thisDemand = routingType_thisLayer == RoutingType.SOURCE_ROUTING ? netPlan.getDemandRoutes(layerId, demandId) : new LinkedHashSet<Long>();

					Pair<Long, Long> nodePair_thisDemand = netPlan.getDemandIngressEgressNodePair(layerId, demandId);
					long ingressNodeId = nodePair_thisDemand.getFirst();
					long egressNodeId = nodePair_thisDemand.getSecond();
					String ingressNodeName = netPlan.getNodeName(ingressNodeId);
					String egressNodeName = netPlan.getNodeName(egressNodeId);
					
					writer.writeStartElement("demand");
					writer.writeAttribute("id", Long.toString(demandId));
					writer.writeAttribute("ingressNodeId", ingressNodeId + (ingressNodeName.isEmpty() ? "" : " (" + ingressNodeName + ")"));
					writer.writeAttribute("egressNodeId", egressNodeId + (egressNodeName.isEmpty() ? "" : " (" + egressNodeName + ")"));
					writer.writeAttribute("coupledTo", coupledLink == null ? "" : String.format("link %d (layer %d)", coupledLink.getSecond(), coupledLink.getFirst()));
					writer.writeAttribute("offeredTraffic", String.format("%.3f", h_d));
					writer.writeAttribute("carriedTraffic", String.format("%.3f", netPlan.getDemandCarriedTraffic(layerId, demandId)));
					writer.writeAttribute("blockedTrafficPercentage", String.format("%.3f", h_d == 0 ? 0 : 100 * lostTraffic_d / h_d));
					writer.writeAttribute("routeIds", routeIds_thisDemand.size() + (routeIds_thisDemand.size() > 0 ? " (" + CollectionUtils.join(routeIds_thisDemand, ",") + ")" : ""));
					writer.writeAttribute("attributes", StringUtils.mapToString(netPlan.getDemandAttributeMap(layerId, demandId)));
					writer.writeEndElement();
				}
				
				if (routingType_thisLayer == RoutingType.SOURCE_ROUTING)
				{
					Map<Long, Double> linkUtilizationMap = netPlan.getLinkUtilizationMap(layerId);
					
					/* Per-route information */
					Set<Long> routeIds_thisLayer = netPlan.getRouteIds(layerId);
					for(long routeId : routeIds_thisLayer)
					{
						long demandId = netPlan.getRouteDemand(layerId, routeId);

						List<Long> seqLinks = netPlan.getRouteSequenceOfLinks(layerId, routeId);
						List<Long> seqNodes = netPlan.getRouteSequenceOfNodes(layerId, routeId);
						Map<Long, Double> utilizationMap = CollectionUtils.selectEntries(linkUtilizationMap, new LinkedHashSet<Long>(seqLinks));
						double maxUtilization = utilizationMap.isEmpty() ? 0 : DoubleUtils.maxValue(utilizationMap);

						Pair<Long, Long> nodePair_thisDemand = netPlan.getDemandIngressEgressNodePair(layerId, demandId);
						long ingressNodeId = nodePair_thisDemand.getFirst();
						long egressNodeId = nodePair_thisDemand.getSecond();
						String ingressNodeName = netPlan.getNodeName(ingressNodeId);
						String egressNodeName = netPlan.getNodeName(egressNodeId);
						
						writer.writeStartElement("route");
						writer.writeAttribute("id", Long.toString(routeId));
						writer.writeAttribute("demandId", Long.toString(demandId));
						writer.writeAttribute("ingressNodeId", ingressNodeId + (ingressNodeName.isEmpty() ? "" : " (" + ingressNodeName + ")"));
						writer.writeAttribute("egressNodeId", egressNodeId + (egressNodeName.isEmpty() ? "" : " (" + egressNodeName + ")"));
						writer.writeAttribute("offeredTraffic", String.format("%.3f", netPlan.getDemandOfferedTraffic(layerId, demandId)));
						writer.writeAttribute("carriedTraffic", String.format("%.3f", netPlan.getRouteCarriedTraffic(layerId, routeId)));
						writer.writeAttribute("occupiedCapacity", String.format("%.3f", netPlan.getRouteOccupiedCapacity(layerId, routeId)));
						writer.writeAttribute("seqLinks", CollectionUtils.join(seqLinks, " => "));
						writer.writeAttribute("seqNodes", CollectionUtils.join(seqNodes, " => "));
						writer.writeAttribute("numHops", Integer.toString(netPlan.getRouteNumberOfHops(layerId, routeId)));
						writer.writeAttribute("routeLength", String.format("%.3f", netPlan.getRouteLengthInKm(layerId, routeId)));
						writer.writeAttribute("routeDelay", String.format("%.3f", 1000 * netPlan.getRoutePropagationDelayInSeconds(layerId, routeId)));
						writer.writeAttribute("maxUtilization", String.format("%.3f", maxUtilization));
						writer.writeAttribute("backupSegmentList", CollectionUtils.join(netPlan.getRouteBackupSegmentList(layerId, routeId), ", "));
						writer.writeAttribute("attributes", StringUtils.mapToString(netPlan.getRouteAttributeMap(layerId, routeId)));
						writer.writeEndElement();
					}
					
					/* Per-protection-segment information */
					Set<Long> segmentIds = netPlan.getProtectionSegmentIds(layerId);
					for(long segmentId : segmentIds)
					{
						List<Long> seqLinks = netPlan.getProtectionSegmentSequenceOfLinks(layerId, segmentId);
						List<Long> seqNodes = netPlan.getProtectionSegmentSequenceOfNodes(layerId, segmentId);
						Set<Long> routeIds_thisSegment = netPlan.getProtectionSegmentRoutes(layerId, segmentId);
						int numRoutes = routeIds_thisSegment.size();
						
						Pair<Long, Long> nodePair_thisSegment = netPlan.getProtectionSegmentOriginDestinationNodePair(layerId, segmentId);
						long originNodeId = nodePair_thisSegment.getFirst();
						long destinationNodeId = nodePair_thisSegment.getSecond();
						String originNodeName = netPlan.getNodeName(originNodeId);
						String destinationNodeName = netPlan.getNodeName(destinationNodeId);
						
						writer.writeStartElement("segment");
						writer.writeAttribute("id", Long.toString(segmentId));
						writer.writeAttribute("originNodeId", originNodeId + (originNodeName.isEmpty() ? "" : " (" + originNodeName + ")"));
						writer.writeAttribute("destinationNodeId", destinationNodeId + (destinationNodeName.isEmpty() ? "" : " (" + destinationNodeName + ")"));
						writer.writeAttribute("reservedCapacity", String.format("%.3f", netPlan.getProtectionSegmentReservedCapacity(layerId, segmentId)));
						writer.writeAttribute("carriedTraffic", String.format("%.3f", netPlan.getProtectionSegmentCarriedTraffic(layerId, segmentId)));
						writer.writeAttribute("seqLinks", CollectionUtils.join(seqLinks, " => "));
						writer.writeAttribute("seqNodes", CollectionUtils.join(seqNodes, " => "));
						writer.writeAttribute("numHops", Integer.toString(netPlan.getProtectionSegmentNumberOfHops(layerId, segmentId)));
						writer.writeAttribute("segmentLength", String.format("%.3f", netPlan.getProtectionSegmentLengthInKm(layerId, segmentId)));
						writer.writeAttribute("segmentDelay", String.format("%.3f", 1000 * netPlan.getProtectionSegmentPropagationDelayInSeconds(layerId, segmentId)));
						writer.writeAttribute("isDedicated",  numRoutes > 1 ? "Shared" : (numRoutes == 0 ? "Not used" : "Dedicated"));
						writer.writeAttribute("routes", numRoutes > 0 ? numRoutes + " (" + CollectionUtils.join(routeIds_thisSegment, ", ") + ")" : "none");
						writer.writeAttribute("attributes", StringUtils.mapToString(netPlan.getProtectionSegmentAttributeMap(layerId, segmentId)));
						writer.writeEndElement();
					}
				}
				else
				{
					/* Per-forwarding rule information */
					Set<Pair<Long, Long>> forwardingRules_thisLayer = new TreeSet<Pair<Long, Long>>(new ForwardingRuleComparator(netPlan, layerId));
					forwardingRules_thisLayer.addAll(netPlan.getForwardingRules(layerId));
					if (!forwardingRules_thisLayer.isEmpty())
					{
						for(Pair<Long, Long> forwardingRule : forwardingRules_thisLayer)
						{
							long demandId = forwardingRule.getFirst();
							long linkId = forwardingRule.getSecond();
							Pair<Long, Long> nodePair_thisDemand = netPlan.getDemandIngressEgressNodePair(layerId, demandId);
							long ingressNodeId = nodePair_thisDemand.getFirst();
							long egressNodeId = nodePair_thisDemand.getSecond();
							String ingressNodeName = netPlan.getNodeName(ingressNodeId);
							String egressNodeName = netPlan.getNodeName(egressNodeId);

							Pair<Long, Long> nodePair_thisLink = netPlan.getLinkOriginDestinationNodePair(layerId, linkId);
							long originNodeId = nodePair_thisLink.getFirst();
							long destinationNodeId = nodePair_thisLink.getSecond();
							String originNodeName = netPlan.getNodeName(nodePair_thisLink.getFirst());
							String destinationNodeName = netPlan.getNodeName(nodePair_thisLink.getSecond());

							Map<String, String> forwardingRuleAttributes = netPlan.getForwardingRuleAttributeMap(layerId, forwardingRule);

							writer.writeStartElement("forwardingRule");
							writer.writeAttribute("nodeId", originNodeId + (originNodeName.isEmpty() ? "" : " (" + originNodeName + ")"));
							writer.writeAttribute("demandId", demandId + " [" + ingressNodeId + (ingressNodeName.isEmpty() ? "" : " (" + ingressNodeName + ")") + " -> " + egressNodeId + (egressNodeName.isEmpty() ? "" : " (" + egressNodeName + ")") + "]");
							writer.writeAttribute("linkId", linkId + " [" + originNodeId + (originNodeName.isEmpty() ? "" : " (" + originNodeName + ")") + " -> " + destinationNodeId + (destinationNodeName.isEmpty() ? "" : " (" + destinationNodeName + ")") + "]");
							writer.writeAttribute("splittingRatio", String.format("%.3f", netPlan.getForwardingRuleSplittingRatio(layerId, forwardingRule)));
							writer.writeAttribute("carriedTraffic", String.format("%.3f", netPlan.getForwardingRuleCarriedTraffic(layerId, forwardingRule)));
							writer.writeAttribute("attributes", forwardingRuleAttributes.isEmpty() ? "none" : StringUtils.mapToString(forwardingRuleAttributes));
							writer.writeEndElement();
						}
					}
				}

				/* Per-SRG information */
				if (!srgIds.isEmpty())
				{
					for (long srgId : srgIds)
					{
						writer.writeStartElement("srg");
						writer.writeAttribute("id", Long.toString(srgId));

						double mttf_thisSRG = netPlan.getSRGMeanTimeToFailInHours(srgId);
						double mttr_thisSRG = netPlan.getSRGMeanTimeToRepairInHours(srgId);
						writer.writeAttribute("mttf", mttf_thisSRG == Double.MAX_VALUE ? "&infin;" : StringUtils.secondsToYearsDaysHoursMinutesSeconds(mttf_thisSRG * 3600));
						writer.writeAttribute("mttr", StringUtils.secondsToYearsDaysHoursMinutesSeconds(mttr_thisSRG * 3600));
						
						Set<Long> nodeIds_thisSRG = netPlan.getSRGNodes(srgId);
						writer.writeAttribute("nodeIds", nodeIds_thisSRG.isEmpty() ? "none" : CollectionUtils.join(nodeIds_thisSRG, ", "));

						Set<Long> linkIds_thisSRG = netPlan.getSRGLinks(layerId, srgId);
						writer.writeAttribute("linkIds", linkIds_thisSRG.isEmpty() ? "none" : CollectionUtils.join(linkIds_thisSRG, ", "));

						if (routingType_thisLayer == RoutingType.SOURCE_ROUTING)
						{
							Set<Long> routeIds_thisSRG = netPlan.getSRGRoutes(layerId, srgId);
							int numRoutes_thisSRG = routeIds_thisSRG.size();
							writer.writeAttribute("routeIds", numRoutes_thisSRG > 0 ? numRoutes_thisSRG + " (" + CollectionUtils.join(routeIds_thisSRG, ", ") + ")" : "none");

							Set<Long> segmentIds_thisSRG = netPlan.getSRGProtectionSegments(layerId, srgId);
							int numSegments_thisSRG = segmentIds_thisSRG.size();
							writer.writeAttribute("segmentIds", numSegments_thisSRG > 0 ? numSegments_thisSRG + " (" + CollectionUtils.join(segmentIds_thisSRG, ", ") + ")" : "none");
						}

						Map<String, String> srgAttributes = netPlan.getSRGAttributeMap(srgId);
						writer.writeAttribute("attributes", srgAttributes.isEmpty() ? "none" : StringUtils.mapToString(netPlan.getSRGAttributeMap(srgId)));
						writer.writeEndElement();
					}
				}

				writer.writeEndElement();
			}

			writer.writeEndElement();
			writer.writeEndDocument();
			writer.flush();
			writer.close();

			String xml = os.toString(StandardCharsets.UTF_8.name());
			return HTMLUtils.getHTMLFromXML(xml, Report_networkDesign.class.getResource("/com/net2plan/examples/reports/networkDesign/Report_networkDesign.xsl").toURI().toURL());
		}
		catch (IOException | URISyntaxException | XMLStreamException ex)
		{
			throw new RuntimeException(ex);
		}
	}

	@Override
	public String getDescription()
	{
		return "This report shows basic information about the network design";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		return null;
	}

	@Override
	public String getTitle()
	{
		return "Network design report";
	}
}